from cwt import COSEKey
from typing import Union

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.x509 import Certificate
from cryptography.hazmat.primitives import hashes, serialization

from pymdoccbor import settings

class MsoX509Fabric:
    """
    MsoX509Fabric helper class to create a new mso
    """

    def selfsigned_x509cert(self, encoding: str = "DER") -> Union[Certificate, bytes]:
        """
        Returns an X.509 certificate derived from the private key of the MSO Issuer

        :param encoding: str: the encoding to use, default is DER

        :return: Union[Certificate, bytes]: the X.509 certificate
        """
        ckey = COSEKey.from_bytes(self.private_key.encode())

        subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, settings.X509_COUNTRY_NAME),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, settings.X509_STATE_OR_PROVINCE_NAME),
            x509.NameAttribute(NameOID.LOCALITY_NAME, settings.X509_LOCALITY_NAME),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, settings.X509_ORGANIZATION_NAME),
            x509.NameAttribute(NameOID.COMMON_NAME, settings.X509_COMMON_NAME),
        ])
        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            ckey.key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            settings.X509_NOT_VALID_BEFORE
        ).not_valid_after(
            settings.X509_NOT_VALID_AFTER
        ).add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.UniformResourceIdentifier(
                        settings.X509_SAN_URL
                    )
                ]
            ),
            critical=False,
            # Sign our certificate with our private key
        ).sign(ckey.key, hashes.SHA256())

        if not encoding:
            return cert
        else:
            return cert.public_bytes(
                getattr(serialization.Encoding, encoding)
            )



import os

PKEY = {
    'KTY': 'EC2',
    'CURVE': 'P_256',
    'ALG': 'ES256',
    'D': b"<\xe5\xbc;\x08\xadF\x1d\xc5\x0czR'T&\xbb\x91\xac\x84\xdc\x9ce\xbf\x0b,\x00\xcb\xdd\xbf\xec\xa2\xa5",
    'KID': b"demo-kid"
}


import cbor2
import datetime
import hashlib
import secrets
import uuid
import logging

logger = logging.getLogger("pymdoccbor")

from pycose.headers import Algorithm #, KID
from pycose.keys import CoseKey, EC2Key
from pycose.messages import Sign1Message

from typing import Union

from pymdoccbor.exceptions import MsoPrivateKeyRequired
from pymdoccbor import settings
from pymdoccbor.x509 import MsoX509Fabric
from pymdoccbor.tools import shuffle_dict
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.x509 import Certificate


from cbor_diag import *


class MsoIssuer(MsoX509Fabric):
    """
    MsoIssuer helper class to create a new mso
    """

    def __init__(
        self,
        data: dict,
        validity: dict,
        cert_path: str = None,
        key_label: str = None,
        user_pin: str = None,
        lib_path: str = None,
        slot_id: int = None,
        kid: str = None,
        alg: str = None,
        hsm: bool = False,
        private_key: Union[dict, CoseKey] = None,
        digest_alg: str = settings.PYMDOC_HASHALG,
        revocation: dict = None
    ) -> None:
        """
        Initialize a new MsoIssuer

        :param data: dict: the data to sign
        :param validity: validity: the validity info of the mso
        :param cert_path: str: the path to the certificate
        :param key_label: str: key label
        :param user_pin: str: user pin
        :param lib_path: str: path to the library cryptographic library
        :param slot_id: int: slot id
        :param kid: str: key id
        :param alg: str: hashig algorithm
        :param hsm: bool: hardware security module
        :param private_key: Union[dict, CoseKey]: the signing key
        :param digest_alg: str: the digest algorithm
        :param revocation: dict: revocation status dict to include in the mso, it may include status_list and identifier_list keys
        """

        if not hsm:
            if private_key:
                if isinstance(private_key, dict):
                    self.private_key = CoseKey.from_dict(private_key)
                    if not self.private_key.kid:
                        self.private_key.kid = str(uuid.uuid4())
                elif isinstance(private_key, CoseKey):
                    self.private_key = private_key
                else:
                    raise ValueError("private_key must be a dict or CoseKey object")
            else:
                raise MsoPrivateKeyRequired("MSO Writer requires a valid private key")

        if not validity:
            raise ValueError("validity must be present")

        if not alg:
            raise ValueError("alg must be present")

        self.data: dict = data
        self.hash_map: dict = {}
        self.cert_path = cert_path
        self.disclosure_map: dict = {}
        self.digest_alg: str = digest_alg
        self.key_label = key_label
        self.user_pin = user_pin
        self.lib_path = lib_path
        self.slot_id = slot_id
        self.hsm = hsm
        self.alg = alg
        self.kid = kid
        self.validity = validity
        self.revocation = revocation

        alg_map = {"ES256": "sha256", "ES384": "sha384", "ES512": "sha512"}

        hashfunc = getattr(hashlib, alg_map.get(self.alg))

        digest_cnt = 0
        for ns, values in data.items():
            self.disclosure_map[ns] = {}
            self.hash_map[ns] = {}
            for k, v in shuffle_dict(values).items():
                _rnd_salt = secrets.token_bytes(settings.DIGEST_SALT_LENGTH)

                _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k, None)

                if _value_cbortag:
                    v = cbor2.CBORTag(_value_cbortag, value=v)

                if isinstance(v, dict):
                    for k2, v2 in v.items():
                        _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k2, None)
                        if _value_cbortag:
                            v[k2] = cbor2.CBORTag(_value_cbortag, value=v2)

                if isinstance(v, list) and k != "nationality":
                    for item in v:
                        for k2, v2 in item.items():
                            _value_cbortag = settings.CBORTAGS_ATTR_MAP.get(k2, None)
                            if _value_cbortag:
                                item[k2] = cbor2.CBORTag(_value_cbortag, value=v2)

                self.disclosure_map[ns][digest_cnt] = cbor2.CBORTag(
                    24,
                    value=cbor2.dumps(
                        {
                            "digestID": digest_cnt,
                            "random": _rnd_salt,
                            "elementIdentifier": k,
                            "elementValue": v,
                        },
                        canonical=True,
                    ),
                )

                self.hash_map[ns][digest_cnt] = hashfunc(
                    cbor2.dumps(self.disclosure_map[ns][digest_cnt], canonical=True)
                ).digest()

                digest_cnt += 1

    def format_datetime_repr(self, dt: datetime.datetime) -> str:
        """
        Format a datetime object to a string

        :param dt: datetime.datetime: the datetime object
        :return: str: the formatted string
        """
        return dt.isoformat().split(".")[0] + "Z"

    def sign(
        self,
        device_key: Union[dict, None] = None,
        valid_from: Union[None, datetime.datetime] = None,
        doctype: str = None,
    ) -> Sign1Message:
        """
        Sign a mso and returns it

        :param device_key: Union[dict, None]: the device key
        :param valid_from: Union[None, datetime.datetime]: the valid from date
        :param doctype: str: the document type

        :return: Sign1Message: the signed mso
        """

        utcnow = datetime.datetime.utcnow()
        valid_from = datetime.datetime.strptime(
            self.validity["issuance_date"], "%Y-%m-%d"
        )

        if settings.PYMDOC_EXP_DELTA_HOURS:
            exp = utcnow + datetime.timedelta(hours=settings.PYMDOC_EXP_DELTA_HOURS)
        else:
            # five years
            exp = datetime.datetime.strptime(self.validity["expiry_date"], "%Y-%m-%d")
            # exp = utcnow + datetime.timedelta(hours=(24 * 365) * 5)

        if utcnow > valid_from:
            valid_from = utcnow

        alg_map = {"ES256": "SHA-256", "ES384": "SHA-384", "ES512": "SHA-512"}

        payload = {
            "docType": doctype or list(self.hash_map)[0],
            "version": "1.0",
            "validityInfo": {
                "signed": cbor2.CBORTag(0, self.format_datetime_repr(utcnow)),
                "validFrom": cbor2.CBORTag(
                    0, self.format_datetime_repr(valid_from or utcnow)
                ),
                "validUntil": cbor2.CBORTag(0, self.format_datetime_repr(exp)),
            },
            "valueDigests": self.hash_map,
            "deviceKeyInfo": {
                "deviceKey": device_key,
            },
            "digestAlgorithm": alg_map.get(self.alg)
        }
        if self.revocation is not None:
            payload.update({"status": self.revocation})

        if self.cert_path:
            # Try to load the certificate file
            with open(self.cert_path, "rb") as file:
                certificate = file.read()
            _parsed_cert: Union[Certificate, None] = None
            try:
                _parsed_cert = x509.load_pem_x509_certificate(certificate)
            except Exception as e:
                logger.error(f"Certificate at {self.cert_path} could not be loaded as PEM, trying DER")
            
            if not _parsed_cert:
                try:
                    _parsed_cert = x509.load_der_x509_certificate(certificate)
                except Exception as e:
                    _err_msg = f"Certificate at {self.cert_path} could not be loaded as DER"
                    logger.error(_err_msg)

            if _parsed_cert:
                cert = _parsed_cert
            else:
                raise Exception(f"Certificate at {self.cert_path} failed parse")
            _cert = cert.public_bytes(getattr(serialization.Encoding, "DER"))
        else:
            _cert = self.selfsigned_x509cert()

        if self.hsm:
            # print("payload diganostic notation: \n",cbor2diag(cbor2.dumps(cbor2.CBORTag(24, cbor2.dumps(payload)))))

            mso = Sign1Message(
                phdr={
                    Algorithm: self.alg,
                    # 33: _cert
                },
                # TODO: x509 (cbor2.CBORTag(33)) and federation trust_chain support (cbor2.CBORTag(27?)) here
                # 33 means x509chain standing to rfc9360
                # in both protected and unprotected for interop purpose .. for now.
                uhdr={33: _cert},
                payload=cbor2.dumps(
                    cbor2.CBORTag(24, cbor2.dumps(payload, canonical=True)),
                    canonical=True,
                ),
            )

        else:
            logger.debug("payload diagnostic notation: {cbor2diag(cbor2.dumps(cbor2.CBORTag(24,cbor2.dumps(payload))))}")

            mso = Sign1Message(
                phdr={
                    Algorithm: self.private_key.alg,
                    # KID: self.private_key.kid,
                    # 33: _cert
                },
                # TODO: x509 (cbor2.CBORTag(33)) and federation trust_chain support (cbor2.CBORTag(27?)) here
                # 33 means x509chain standing to rfc9360
                # in both protected and unprotected for interop purpose .. for now.
                uhdr={33: _cert},
                payload=cbor2.dumps(
                    cbor2.CBORTag(24, cbor2.dumps(payload, canonical=True)),
                    canonical=True,
                ),
            )

            mso.key = self.private_key

        return mso



import base64
import binascii
import cbor2
import logging
from datetime import datetime, timezone
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from pycose.keys import CoseKey, EC2Key
from typing import Union

from pymdoccbor.mso.issuer import MsoIssuer

from cbor_diag import cbor2diag


logger = logging.getLogger("pymdoccbor")


class MdocCborIssuer:
    """
    MdocCborIssuer helper class to create a new mdoc
    """
    def __init__(
        self,
        key_label: str = None,
        user_pin: str = None,
        lib_path: str = None,
        slot_id: int = None,
        hsm: bool = False,
        alg: str = None,
        kid: str = None,
        private_key: Union[dict, CoseKey] = {},
    ):
        """
        Initialize a new MdocCborIssuer

        :param key_label: str: key label
        :param user_pin: str: user pin
        :param lib_path: str: path to the library cryptographic library
        :param slot_id: int: slot id
        :param hsm: bool: hardware security module
        :param alg: str: hashig algorithm
        :param kid: str: key id
        :param private_key: Union[dict, CoseKey]: private key
        """
        self.version: str = "1.0"
        self.status: int = 0

        if private_key:
            if isinstance(private_key, dict):
                self.private_key = CoseKey.from_dict(private_key)
            elif isinstance(private_key, EC2Key):
                ec2_encoded = private_key.encode()
                ec2_decoded = CoseKey.decode(ec2_encoded)
                self.private_key = ec2_decoded
            elif isinstance(private_key, CoseKey):
                self.private_key = private_key
            else:
                raise ValueError("private_key must be a dict or CoseKey object")

        self.signed: dict = {}
        self.key_label = key_label
        self.user_pin = user_pin
        self.lib_path = lib_path
        self.slot_id = slot_id
        self.hsm = hsm
        self.alg = alg
        self.kid = kid

    def new(
        self,
        data: dict,
        doctype: str,
        validity: dict = None,
        devicekeyinfo: Union[dict, CoseKey, str] = None,
        cert_path: str = None,
        revocation: dict = None
    ) -> dict:
        """
        create a new mdoc with signed mso

        :param data: dict: data to be signed
        :param doctype: str: document type
        :param validity: dict: validity info
        :param devicekeyinfo: Union[dict, CoseKey, str]: device key info
        :param cert_path: str: path to the certificate
        :param revocation: dict: revocation status dict it may include status_list and identifier_list keys

        :return: dict: signed mdoc
        """
        if isinstance(devicekeyinfo, dict):
            devicekeyinfoCoseKeyObject = CoseKey.from_dict(devicekeyinfo)
            devicekeyinfo = {
                1: devicekeyinfoCoseKeyObject.kty.identifier,
                -1: devicekeyinfoCoseKeyObject.crv.identifier,
                -2: devicekeyinfoCoseKeyObject.x,
                -3: devicekeyinfoCoseKeyObject.y,
            }
        if isinstance(devicekeyinfo, str):
            device_key_bytes = base64.urlsafe_b64decode(devicekeyinfo.encode("utf-8"))
            public_key:EllipticCurvePublicKey = serialization.load_pem_public_key(device_key_bytes)
            curve_name = public_key.curve.name
            curve_map = {
                "secp256r1": 1,  # NIST P-256
                "secp384r1": 2,  # NIST P-384
                "secp521r1": 3,  # NIST P-521
                "brainpoolP256r1": 8,  # Brainpool P-256
                "brainpoolP384r1": 9,  # Brainpool P-384
                "brainpoolP512r1": 10,  # Brainpool P-512
                # Add more curve mappings as needed
            }
            curve_identifier = curve_map.get(curve_name)

            # Extract the x and y coordinates from the public key
            x = public_key.public_numbers().x.to_bytes(
                (public_key.public_numbers().x.bit_length() + 7)
                // 8,  # Number of bytes needed
                "big",  # Byte order
            )

            y = public_key.public_numbers().y.to_bytes(
                (public_key.public_numbers().y.bit_length() + 7)
                // 8,  # Number of bytes needed
                "big",  # Byte order
            )

            devicekeyinfo = {
                1: 2,
                -1: curve_identifier,
                -2: x,
                -3: y,
            }

        else:
            devicekeyinfo: CoseKey = devicekeyinfo

        if self.hsm:
            msoi = MsoIssuer(
                data=data,
                cert_path=cert_path,
                hsm=self.hsm,
                key_label=self.key_label,
                user_pin=self.user_pin,
                lib_path=self.lib_path,
                slot_id=self.slot_id,
                alg=self.alg,
                kid=self.kid,
                validity=validity,
                revocation=revocation
            )

        else:
            msoi = MsoIssuer(
                data=data,
                private_key=self.private_key,
                alg=self.alg,
                cert_path=cert_path,
                validity=validity,
                revocation=revocation
            )

        mso = msoi.sign(doctype=doctype, device_key=devicekeyinfo,valid_from=datetime.now(timezone.utc))

        mso_cbor = mso.encode(
            tag=False,
            hsm=self.hsm,
            key_label=self.key_label,
            user_pin=self.user_pin,
            lib_path=self.lib_path,
            slot_id=self.slot_id,
        )


        res = {
            "version": self.version,
            "documents": [
                {
                "docType": doctype,  # 'org.iso.18013.5.1.mDL'
                "issuerSigned": {
                    "nameSpaces": {
                        ns: [v for k, v in dgst.items()]
                        for ns, dgst in msoi.disclosure_map.items()
                        },
                    "issuerAuth": cbor2.decoder.loads(mso_cbor),
                    },
                }
            ],
            "status": self.status,
        }

        logger.debug(f"MSO diagnostic notation: {cbor2diag(mso_cbor)}")

        self.signed = res
        return self.signed

    def dump(self) -> bytes:
        """
        Returns the CBOR representation of the signed mdoc

        :return: bytes: CBOR representation of the signed mdoc
        """
        return cbor2.dumps(self.signed, canonical=True)

    def dumps(self) -> bytes:
        """
        Returns the AF binary representation of the signed mdoc

        :return: bytes: AF binary representation of the signed mdoc
        """
        return binascii.hexlify(cbor2.dumps(self.signed, canonical=True))



import cbor2
import cryptography
import logging

from pycose.keys import CoseKey, EC2Key
from pycose.messages import Sign1Message

from typing import Union

from pymdoccbor.exceptions import (
    MsoX509ChainNotFound,
    UnsupportedMsoDataFormat
)
from pymdoccbor import settings
from pymdoccbor.tools import bytes2CoseSign1, cborlist2CoseSign1


logger = logging.getLogger("pymdoccbor")


class MsoVerifier:
    """
    Parameters
        data: CBOR TAG 24

    Example:
        MsoParser(mdoc['documents'][0]['issuerSigned']['issuerAuth'])

    Note
        The signature is contained in an untagged COSE_Sign1
        structure as defined in RFC 8152.
    """

    def __init__(self, data: Union[cbor2.CBORTag, bytes, list]) -> None:
        """
        Initialize the MsoParser object

        :param data: Union[cbor2.CBORTag, bytes, list]: the data to parse
        """

        self._data = data

        # not used
        if isinstance(self._data, bytes):
            self.object: Sign1Message = bytes2CoseSign1(
                cbor2.dumps(cbor2.CBORTag(18, value=self._data)))
        elif isinstance(self._data, list):
            self.object: Sign1Message = cborlist2CoseSign1(self._data)
        else:
            raise UnsupportedMsoDataFormat(
                f"MsoParser only supports raw bytes and list, a {type(data)} was provided"
            )

        self.object.key = None
        self.public_key: cryptography.hazmat.backends.openssl.ec._EllipticCurvePublicKey = None
        self.x509_certificates: list = []

    @property
    def payload_as_cbor(self) -> dict:
        """
        It returns the payload as a CBOR TAG

        :return: dict: the payload as a CBOR TAG 24
        """
        return cbor2.loads(self.object.payload)

    @property
    def payload_as_raw(self):
        return self.object.payload

    @property
    def payload_as_dict(self):
        return cbor2.loads(
            cbor2.loads(self.object.payload).value
        )

    @property
    def raw_public_keys(self) -> bytes:
        """
            it returns the public key extract from x509 certificates
            looking to both phdr and uhdr
        """
        _mixed_heads = self.object.phdr.items() | self.object.uhdr.items()
        for h, v in _mixed_heads:
            if h.identifier == 33:
                return list(self.object.uhdr.values())

        raise MsoX509ChainNotFound(
            "I can't find any valid X509certs, identified by label number 33, "
            "in this MSO."
        )

    def attest_public_key(self):
        logger.warning(
            "TODO: in next releases. "
            "The certificate is to be considered as untrusted, this release "
            "doesn't validate x.509 certificate chain. See next releases and "
            "python certvalidator or cryptography for that."
        )

    def load_public_key(self) -> None:
        """
        Load the public key from the x509 certificate

        :return: None
        """
        self.attest_public_key()

        for i in self.raw_public_keys:
            self.x509_certificates.append(
                cryptography.x509.load_der_x509_certificate(i)
            )

        self.public_key = self.x509_certificates[0].public_key()

        key = EC2Key(
            crv=settings.COSEKEY_HAZMAT_CRV_MAP[self.public_key.curve.name],
            x=self.public_key.public_numbers().x.to_bytes(
                settings.CRV_LEN_MAP[self.public_key.curve.name], 'big'
            ),
            y=self.public_key.public_numbers().y.to_bytes( settings.CRV_LEN_MAP[self.public_key.curve.name], 'big')
        )
        self.object.key = key

    def verify_signature(self) -> bool:
        """"
        Verify the signature of the MSO

        :return: bool: True if the signature is valid, False otherwise
        """
        if not self.object.key:
            self.load_public_key()

        return self.object.verify_signature()



import os
import cbor2
from pymdoccbor.mdoc.issuer import MdocCborIssuer
from pymdoccbor.tests.micov_data import MICOV_DATA
from pymdoccbor.mdoc.verifier import MdocCbor
from pymdoccbor.tests.pkey import PKEY

def test_mdoc_cbor_creation():
    mdoci = MdocCborIssuer(
        private_key=PKEY,
        alg="ES256",
    )
    mdoc = mdoci.new(
        data=MICOV_DATA,
        #devicekeyinfo=PKEY,  # TODO
        doctype="org.micov.medical.1",
        validity={
            "issuance_date": "2024-12-31",
            "expiry_date": "2050-12-31"
        },
    )

    data = cbor2.dumps(mdoc)  

    mdocp = MdocCbor()
    mdocp.loads(data)
    mdocp.verify()

    assert mdoc
    assert 'org.micov.medical.1' in mdocp.disclosure_map
    assert mdocp.disclosure_map == MICOV_DATA


import datetime
import os

from datetime import timezone

COSEKEY_HAZMAT_CRV_MAP = {
    "secp256r1": "P_256",
    "secp384r1": "P_384",
    "secp521r1":  "P_521"
}

CRV_LEN_MAP = {
    "secp256r1": 32,
}

PYMDOC_HASHALG: str = os.getenv("PYMDOC_HASHALG", "SHA-256")
PYMDOC_EXP_DELTA_HOURS: int = os.getenv("PYMDOC_EXP_DELTA_HOURS", 0)

HASHALG_MAP = {
    "SHA-256": "sha256",
    "SHA-512": "sha512",
}

DIGEST_SALT_LENGTH = 32


X509_DER_CERT = os.getenv("X509_DER_CERT", None)

# OR

X509_COUNTRY_NAME           = os.getenv('X509_COUNTRY_NAME', "US")
X509_STATE_OR_PROVINCE_NAME = os.getenv('X509_STATE_OR_PROVINCE_NAME', "California")
X509_LOCALITY_NAME          = os.getenv('X509_LOCALITY_NAME', "San Francisco")
X509_ORGANIZATION_NAME      = os.getenv('X509_ORGANIZATION_NAME', "My Company")
X509_COMMON_NAME            = os.getenv('X509_COMMON_NAME', "mysite.com")

X509_NOT_VALID_BEFORE       = os.getenv('X509_NOT_VALID_BEFORE', datetime.datetime.now(timezone.utc))
X509_NOT_VALID_AFTER_DAYS   = os.getenv('X509_NOT_VALID_AFTER_DAYS', 10)
X509_NOT_VALID_AFTER        = os.getenv(
    'X509_NOT_VALID_AFTER', 
    datetime.datetime.now(timezone.utc) + datetime.timedelta(
        days=X509_NOT_VALID_AFTER_DAYS
    )
)

X509_SAN_URL = os.getenv(
    "X509_SAN_URL", "https://credential-issuer.example.org"
)

CBORTAGS_ATTR_MAP = {
    "birth_date": 1004,
    "expiry_date": 1004,
    "issue_date": 1004,
    "issuance_date": 1004,
}
